@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_ctrls_admin -- a module implementing admin control handlers
  *
- * Copyright (c) 2000-2005 TJ Saunders
+ * Copyright (c) 2000-2010 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,18 +25,18 @@
  * This is mod_controls, contrib software for proftpd 1.2 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_ctrls_admin.c,v 1.23 2005/04/30 19:59:27 castaglia Exp $
+ * $Id: mod_ctrls_admin.c,v 1.40 2010/01/10 20:01:30 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 #include "mod_ctrls.h"
 
-#define MOD_CTRLS_ADMIN_VERSION		"mod_ctrls_admin/0.9.3"
+#define MOD_CTRLS_ADMIN_VERSION		"mod_ctrls_admin/0.9.6"
 
 /* Make sure the version of proftpd is as necessary. */
-#if PROFTPD_VERSION_NUMBER < 0x0001021001
-# error "ProFTPD 1.2.10rc1 or later required"
+#if PROFTPD_VERSION_NUMBER < 0x0001030001
+# error "ProFTPD 1.3.0rc1 or later required"
 #endif
 
 #ifndef PR_USE_CTRLS
@@ -62,25 +62,12 @@ static ctrls_acttab_t ctrls_admin_acttab[];
 /* Pool for this module's use */
 static pool *ctrls_admin_pool = NULL;
 
-/* For the 'dump' action */
-static pr_ctrls_t *ctrls_dump_ctrl = NULL;
+static unsigned int ctrls_admin_nrestarts = 0;
+static time_t ctrls_admin_start = 0;
 
 /* Support routines
  */
 
-static void ctrls_admin_printf(const char *fmt, ...) {
-  char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
-  va_list msg;
-
-  va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
-  va_end(msg);
-
-  buf[sizeof(buf)-1] = '\0';
-
-  pr_ctrls_add_response(ctrls_dump_ctrl, "%s", buf);
-}
-
 #if 0
 /* Will be used when scheduled shutdowns are supported.. */
 static unsigned char isnumeric(char *str) {
@@ -103,6 +90,25 @@ static int respcmp(const void *a, const void *b) {
   return strcmp(*((char **) a), *((char **) b));
 }
 
+#ifdef PR_USE_DEVEL
+static pr_ctrls_t *mem_ctrl = NULL;
+
+static void mem_printf(const char *fmt, ...) {
+  char buf[PR_TUNABLE_BUFFER_SIZE];
+  va_list msg;
+
+  memset(buf, '\0', sizeof(buf)); 
+
+  va_start(msg, fmt);
+  vsnprintf(buf, sizeof(buf), fmt, msg);
+  va_end(msg);
+ 
+  buf[sizeof(buf)-1] = '\0';
+ 
+  pr_ctrls_add_response(mem_ctrl, "pool: %s", buf);
+}
+#endif /* !PR_USE_DEVEL */
+
 /* Controls handlers
  */
 
@@ -110,7 +116,7 @@ static int ctrls_handle_debug(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
 
   /* Check the debug ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "debug")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "debug")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -138,9 +144,27 @@ static int ctrls_handle_debug(pr_ctrls_t *ctrl, int reqargc,
     }
   
     pr_log_setdebuglevel(level);
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: level set to %d", level);
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: level set to %d", level);
     pr_ctrls_add_response(ctrl, "debug level set to %d", level);
 
+#ifdef PR_USE_DEVEL
+  /* Handle 'debug memory' requests */
+  } else if (strcmp(reqargv[0], "mem") == 0 ||
+             strcmp(reqargv[0], "memory") == 0) {
+
+    if (reqargc != 1) {
+      pr_ctrls_add_response(ctrl, "debug: too many parameters");
+      return -1;
+    }
+
+    mem_ctrl = ctrl;
+    pr_pool_debug_memory(mem_printf);
+    mem_ctrl = NULL;
+
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: dumped memory info");
+
+#endif /* PR_USE_DEVEL */
+
   } else {
     pr_ctrls_add_response(ctrl, "unknown debug action: '%s'", reqargv[0]);
     return -1;
@@ -154,7 +178,7 @@ static int ctrls_handle_dns(pr_ctrls_t *ctrl, int reqargc,
   int bool;
 
   /* Check the dns ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "dns")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "dns")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -167,32 +191,46 @@ static int ctrls_handle_dns(pr_ctrls_t *ctrl, int reqargc,
     return -1;
   }
 
-  if (reqargc != 1) {
+  if (reqargc != 1 &&
+      reqargc != 2) {
     pr_ctrls_add_response(ctrl, "dns: wrong number of parameters");
     return -1;
   }
 
-  bool = pr_is_boolean(reqargv[0]);
-  if (bool == -1) {
-    pr_ctrls_add_response(ctrl, "dns: error: expected Boolean parameter: '%s'",
-      reqargv[0]);
-    return -1;
+  if (reqargc == 2 &&
+      strcmp(reqargv[0], "cache") == 0) {
+    if (strcmp(reqargv[1], "clear") != 0) {
+      pr_ctrls_add_response(ctrl,
+        "dns: error: expected 'clear' command: '%s'", reqargv[1]);
+      return -1;
+    }
+
+    pr_netaddr_clear_cache();
+    pr_ctrls_add_response(ctrl, "dns: netaddr cache cleared");
+    
+  } else {
+    bool = pr_str_is_boolean(reqargv[0]);
+    if (bool == -1) {
+      pr_ctrls_add_response(ctrl,
+        "dns: error: expected Boolean parameter: '%s'", reqargv[0]);
+      return -1;
+    }
+
+    ServerUseReverseDNS = bool;
+
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "dns: UseReverseDNS set to '%s'",
+      bool ? "on" : "off");
+    pr_ctrls_add_response(ctrl, "dns: UseReverseDNS set to '%s'",
+      bool ? "on" : "off");
   }
 
-  ServerUseReverseDNS = bool;
-
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "dns: UseReverseDNS set to '%s'",
-    bool ? "on" : "off");
-  pr_ctrls_add_response(ctrl, "dns: UseReverseDNS set to '%s'",
-    bool ? "on" : "off");
-
   return 0;
 }
 
 static int admin_addr_down(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
     unsigned int port) {
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "down: disabling %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "down: disabling %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   if (pr_ipbind_close(addr, port, FALSE) < 0) {
@@ -219,7 +257,7 @@ static int ctrls_handle_down(pr_ctrls_t *ctrl, int reqargc,
    */
 
   /* Check the 'down' ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "down")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "down")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -364,7 +402,7 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
   int res = 0;
 
   /* Check the kick ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "kick")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "kick")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -393,31 +431,37 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
     for (i = 1; i < reqargc; i++) {
       unsigned char kicked_user = FALSE;
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+      }
 
-      while ((score = pr_scoreboard_read_entry()) != NULL) {
+      while ((score = pr_scoreboard_entry_read()) != NULL) {
         if (strcmp(reqargv[i], score->sce_user) == 0) {
           res = 0;
 
           PRIVS_ROOT
-          res = kill(score->sce_pid, SIGTERM);
+          res = pr_scoreboard_entry_kill(score, SIGTERM);
           PRIVS_RELINQUISH
 
-          if (res == 0)
+          if (res == 0) {
             kicked_user = TRUE;
 
-          else
-            ctrls_log("error kicking user '%s': %s", reqargv[i],
-              strerror(errno));
+          } else {
+            pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+              "error kicking user '%s': %s", reqargv[i], strerror(errno));
+          }
         }
       }
-      if (pr_restore_scoreboard() < 0)
-        ctrls_log("error restoring scoreboard: %s", strerror(errno));
+
+      if (pr_restore_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error restoring scoreboard: %s",
+          strerror(errno));
+      }
 
       if (kicked_user) {
         pr_ctrls_add_response(ctrl, "kicked user '%s'", reqargv[i]);
-        ctrls_log("kicked user '%s'", reqargv[i]);
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked user '%s'", reqargv[i]);
         pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked user '%s'",
           reqargv[i]);
 
@@ -454,13 +498,15 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
 
       addr = pr_netaddr_get_ipstr(na);
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+      }
 
-      while ((score = pr_scoreboard_read_entry()) != NULL) {
+      while ((score = pr_scoreboard_entry_read()) != NULL) {
         if (strcmp(score->sce_client_addr, addr) == 0) {
           PRIVS_ROOT
-          if (kill(score->sce_pid, SIGTERM) == 0)
+          if (pr_scoreboard_entry_kill(score, SIGTERM) == 0)
             kicked_host = TRUE;
           PRIVS_RELINQUISH
         }
@@ -469,7 +515,7 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
 
       if (kicked_host) {
         pr_ctrls_add_response(ctrl, "kicked host '%s'", addr);
-        ctrls_log("kicked host '%s'", addr);
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked host '%s'", addr);
         pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked host '%s'",
           addr);
 
@@ -493,31 +539,37 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
     for (i = 1; i < reqargc; i++) {
       unsigned char kicked_class = FALSE;
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+      }
 
-      while ((score = pr_scoreboard_read_entry()) != NULL) {
+      while ((score = pr_scoreboard_entry_read()) != NULL) {
         if (strcmp(reqargv[i], score->sce_class) == 0) {
           res = 0;
 
           PRIVS_ROOT
-          res = kill(score->sce_pid, SIGTERM);
+          res = pr_scoreboard_entry_kill(score, SIGTERM);
           PRIVS_RELINQUISH
 
-          if (res == 0)
+          if (res == 0) {
             kicked_class = TRUE;
 
-          else
-            ctrls_log("error kicking class '%s': %s", reqargv[i],
-              strerror(errno));
+          } else {
+            pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+              "error kicking class '%s': %s", reqargv[i], strerror(errno));
+          }
         }
       }
-      if (pr_restore_scoreboard() < 0)
-        ctrls_log("error restoring scoreboard: %s", strerror(errno));
+
+      if (pr_restore_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error restoring scoreboard: %s",
+          strerror(errno));
+      }
 
       if (kicked_class) {
         pr_ctrls_add_response(ctrl, "kicked class '%s'", reqargv[i]);
-        ctrls_log("kicked class '%s'", reqargv[i]);
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked class '%s'", reqargv[i]);
         pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked class '%s'",
           reqargv[i]);
 
@@ -538,7 +590,7 @@ static int ctrls_handle_restart(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
 
   /* Check the restart ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "restart")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "restart")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -546,19 +598,66 @@ static int ctrls_handle_restart(pr_ctrls_t *ctrl, int reqargc,
   }
 
   /* Be pedantic */
-  if (reqargc != 0) {
+  if (reqargc > 1) {
     pr_ctrls_add_response(ctrl, "bad number of arguments");
     return -1;
   }
 
-  PRIVS_ROOT
-  raise(SIGHUP);
-  PRIVS_RELINQUISH
+  if (reqargc == 0) {
+    PRIVS_ROOT
+    raise(SIGHUP);
+    PRIVS_RELINQUISH
+
+    pr_ctrls_add_response(ctrl, "restarted server");
+
+  } else if (reqargc == 1) {
+    if (strcmp(reqargv[0], "count") == 0) {
+      struct tm *tm;
+
+      tm = pr_gmtime(ctrl->ctrls_tmp_pool, &ctrls_admin_start);
+      pr_ctrls_add_response(ctrl,
+        "server restarted %u %s since %04d-%02d-%02d %02d:%02d:%02d GMT",
+        ctrls_admin_nrestarts, ctrls_admin_nrestarts != 1 ? "times" : "time",
+        tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min,
+        tm->tm_sec);
+
+    } else {
+      pr_ctrls_add_response(ctrl, "unsupported parameter '%s'", reqargv[0]);
+      return -1;
+    }
+  }
 
-  pr_ctrls_add_response(ctrl, "restarted server");
   return 0;
 }
 
+static int ctrls_handle_scoreboard(pr_ctrls_t *ctrl, int reqargc,
+    char **reqargv) {
+
+  /* Check the scoreboard ACL. */
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "scoreboard")) {
+
+    /* Access denied. */
+    pr_ctrls_add_response(ctrl, "access denied");
+    return -1;
+  }
+
+  if (reqargc != 1) {
+    pr_ctrls_add_response(ctrl, "bad number of arguments");
+    return -1;
+  }
+
+  if (strcmp(reqargv[0], "clean") == 0 ||
+      strcmp(reqargv[0], "scrub") == 0) {
+
+    pr_scoreboard_scrub();
+    pr_ctrls_add_response(ctrl, "scrubbed scoreboard");
+    return 0;
+  }
+
+  pr_ctrls_add_response(ctrl, "unknown scoreboard action '%s'", reqargv[0]);
+  return -1;
+}
+
 static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
   register unsigned int i = 0;
@@ -566,7 +665,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
   char **respargv = NULL;
 
   /* Check the shutdown ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "shutdown")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "shutdown")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -587,7 +686,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
       timeout = atoi(reqargv[1]);
       time(&now);
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
         "shutdown: waiting %u seconds before shutting down", timeout);
 
       /* If the timeout is less than the waiting period, reduce the
@@ -610,9 +709,9 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
       if (timeout &&
           time(NULL) - now > timeout) {
 
-        ctrls_log(MOD_CTRLS_ADMIN_VERSION,
-          "shutdown: %u seconds elapsed, ending remaining sessions",
-          timeout);
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+          "shutdown: %u seconds elapsed, ending %lu remaining sessions",
+          timeout, nkids);
 
         /* End all remaining sessions at this point. */
         PRIVS_ROOT
@@ -622,7 +721,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
         break;
       }
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
         "shutdown: waiting for %lu sessions to end", nkids);
       sleep(waiting);
 
@@ -646,17 +745,18 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
   /* Manually tweak the return value, for the benefit of the client */
   ctrl->ctrls_cb_retval = 0;
 
-  if (pr_ctrls_flush_response(ctrl) < 0)
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+  if (pr_ctrls_flush_response(ctrl) < 0) {
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
       "shutdown: error flushing response: %s", strerror(errno));
+  }
 
   /* For logging/accounting purposes */
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
     "shutdown: flushed to %s/%s client: return value: 0",
     ctrl->ctrls_cl->cl_user, ctrl->ctrls_cl->cl_group);
 
   for (i = 0; i < respargc; i++)
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
       "shutdown: flushed to %s/%s client: '%s'",
       ctrl->ctrls_cl->cl_user, ctrl->ctrls_cl->cl_group, respargv[i]);
 
@@ -670,7 +770,7 @@ static int admin_addr_status(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
     unsigned int port) {
   pr_ipbind_t *ipbind = NULL;
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   /* Fetch the ipbind associated with this address/port. */
@@ -693,7 +793,7 @@ static int ctrls_handle_status(pr_ctrls_t *ctrl, int reqargc,
   register unsigned int i = 0;
 
   /* Check the status ACL. */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "status")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "status")) {
 
     /* Access denied. */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -716,7 +816,7 @@ static int ctrls_handle_status(pr_ctrls_t *ctrl, int reqargc,
     if (strcasecmp(server_str, "all") == 0) {
       pr_ipbind_t *ipbind = NULL;
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking all servers");
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking all servers");
 
       while ((ipbind = pr_ipbind_get(ipbind)) != NULL) {
         const char *ipbind_str = pr_netaddr_get_ipstr(ipbind->ib_addr); 
@@ -757,6 +857,89 @@ static int ctrls_handle_status(pr_ctrls_t *ctrl, int reqargc,
   return 0;
 }
 
+static int ctrls_handle_trace(pr_ctrls_t *ctrl, int reqargc,
+    char **reqargv) {
+#ifdef PR_USE_TRACE
+
+  /* Check the trace ACL. */
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "trace")) {
+
+    /* Access denied. */
+    pr_ctrls_add_response(ctrl, "access denied");
+    return -1;
+  }
+
+  /* Sanity check */
+  if (reqargc < 1 || reqargv == NULL) {
+    pr_ctrls_add_response(ctrl, "trace: missing required parameters");
+    return -1;
+  }
+
+  if (strcmp(reqargv[0], "info") != 0) {
+    register unsigned int i;
+
+    for (i = 0; i < reqargc; i++) {
+      char *channel, *tmp;
+      int level;
+
+      tmp = strchr(reqargv[i], ':');
+      if (!tmp) {
+        pr_ctrls_add_response(ctrl, "trace: badly formatted parameter: '%s'",
+          reqargv[i]);
+        return -1;
+      }
+
+      channel = reqargv[i];
+      *tmp = '\0';
+      level = atoi(++tmp);
+
+      if (pr_trace_set_level(channel, level) < 0) {
+        pr_ctrls_add_response(ctrl,
+          "trace: error setting channel '%s' to level %d: %s", channel, level,
+          strerror(errno));
+        return -1;
+
+      } else {
+        pr_ctrls_add_response(ctrl, "trace: set channel '%s' to level %d",
+          channel, level);
+      }
+    }
+ 
+  } else {
+    pr_table_t *trace_tab = pr_trace_get_table();
+
+    if (trace_tab) {
+      void *key = NULL, *value = NULL;
+
+      pr_ctrls_add_response(ctrl, "%-10s %-6s", "Channel", "Level");
+      pr_ctrls_add_response(ctrl, "---------- ------");
+
+      pr_table_rewind(trace_tab);
+      key = pr_table_next(trace_tab);
+      while (key) {
+        pr_signals_handle();
+
+        value = pr_table_get(trace_tab, (const char *) key, NULL);
+        if (value) {
+          pr_ctrls_add_response(ctrl, "%10s %-6d", (const char *) key,
+            *((int *) value));
+        }
+
+        key = pr_table_next(trace_tab);
+      }
+
+    } else {
+      pr_ctrls_add_response(ctrl, "trace: no info available");
+    }
+  }
+ 
+  return 0;
+#else
+  pr_ctrls_add_response(ctrl, "trace: requires trace support (--enable-trace");
+  return -1;
+#endif /* PR_USE_TRACE */
+}
+
 static int admin_addr_up(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
     unsigned int port) {
   pr_ipbind_t *ipbind = NULL;
@@ -786,12 +969,12 @@ static int admin_addr_up(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
    */
   if (ipbind->ib_server->ServerPort && !ipbind->ib_server->listen) {
     ipbind->ib_server->listen =
-      pr_inet_create_connection(ipbind->ib_server->pool, server_list, -1,
+      pr_inet_create_conn(ipbind->ib_server->pool, server_list, -1,
       (SocketBindTight ? ipbind->ib_server->addr : NULL),
       ipbind->ib_server->ServerPort, FALSE);
   }
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "up: attempting to enable %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "up: attempting to enable %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   PR_OPEN_IPBIND(ipbind->ib_server->addr, ipbind->ib_server->ServerPort,
@@ -818,7 +1001,7 @@ static int ctrls_handle_up(pr_ctrls_t *ctrl, int reqargc,
    */
 
   /* Check the 'up' ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "up")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "up")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -891,13 +1074,13 @@ MODRET set_adminctrlsacls(cmd_rec *cmd) {
       strcmp(cmd->argv[3], "group") != 0)
     CONF_ERROR(cmd, "third parameter must be 'user' or 'group'");
 
-  if ((bad_action = ctrls_set_module_acls(ctrls_admin_acttab,
-      ctrls_admin_pool, actions, cmd->argv[2], cmd->argv[3],
-      cmd->argv[4])) != NULL)
+  bad_action = pr_ctrls_set_module_acls(ctrls_admin_acttab, ctrls_admin_pool,
+    actions, cmd->argv[2], cmd->argv[3], cmd->argv[4]);
+  if (bad_action != NULL)
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown action: '",
       bad_action, "'", NULL));
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: AdminControlsEngine on|off|actions */
@@ -934,13 +1117,14 @@ MODRET set_adminctrlsengine(cmd_rec *cmd) {
      */
     char **actions = ctrls_parse_acl(cmd->tmp_pool, cmd->argv[1]);
 
-    if ((bad_action = ctrls_unregister_module_actions(ctrls_admin_acttab,
-        actions, &ctrls_admin_module)) != NULL)   
+    bad_action = pr_ctrls_unregister_module_actions(ctrls_admin_acttab, actions,
+      &ctrls_admin_module);
+    if (bad_action != NULL)
       CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown action: '",
           bad_action, "'", NULL));
   }
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Event handlers
@@ -982,9 +1166,10 @@ static void ctrls_admin_restart_ev(const void *event_data, void *user_data) {
     /* Allocate and initialize the ACL for this control. */
     ctrls_admin_acttab[i].act_acl = pcalloc(ctrls_admin_pool,
       sizeof(ctrls_acl_t));
-    ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
+    pr_ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
   }
 
+  ctrls_admin_nrestarts++;
   return;
 }
 
@@ -1035,7 +1220,7 @@ static int ctrls_admin_init(void) {
     /* Allocate and initialize the ACL for this control. */
     ctrls_admin_acttab[i].act_acl = pcalloc(ctrls_admin_pool,
       sizeof(ctrls_acl_t));
-    ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
+    pr_ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
 
     if (pr_ctrls_register(&ctrls_admin_module,
         ctrls_admin_acttab[i].act_action, ctrls_admin_acttab[i].act_desc,
@@ -1054,6 +1239,7 @@ static int ctrls_admin_init(void) {
   pr_event_register(&ctrls_admin_module, "core.startup",
     ctrls_admin_startup_ev, NULL);
 
+  time(&ctrls_admin_start);
   return 0;
 }
 
@@ -1070,10 +1256,14 @@ static ctrls_acttab_t ctrls_admin_acttab[] = {
     ctrls_handle_kick },
   { "restart",  "restart the daemon (similar to using HUP)",	NULL,
     ctrls_handle_restart },
+  { "scoreboard", "clean the ScoreboardFile", NULL,
+    ctrls_handle_scoreboard },
   { "shutdown", "shutdown the daemon",	NULL,
     ctrls_handle_shutdown },
   { "status",	"display status of servers",		NULL,
     ctrls_handle_status },
+  { "trace",	"set trace levels",		NULL,
+    ctrls_handle_trace },
   { "up",       "enable a downed virtual server",       NULL,
     ctrls_handle_up },
   { NULL, NULL,	NULL, NULL }
@@ -1110,5 +1300,8 @@ module ctrls_admin_module = {
   ctrls_admin_init,
 
   /* Session initialization function */
-  NULL
+  NULL,
+
+  /* Module version */
+  MOD_CTRLS_ADMIN_VERSION
 };
