@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_sql_sqlite -- Support for connecting to SQLite databases
  *
- * Copyright (c) 2004-2008 TJ Saunders
+ * Copyright (c) 2004-2010 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,11 +21,11 @@
  * with OpenSSL, and distribute the resulting executable, without including
  * the source code for OpenSSL in the source distribution.
  *
- * $Id: mod_sql_sqlite.c,v 1.11 2008/11/03 19:27:24 castaglia Exp $
+ * $Id: mod_sql_sqlite.c,v 1.16.2.1 2010/06/23 14:41:46 castaglia Exp $
  * $Libraries: -lsqlite3 $
  */
 
-#define MOD_SQL_SQLITE_VERSION		"mod_sql_sqlite/0.3"
+#define MOD_SQL_SQLITE_VERSION		"mod_sql_sqlite/0.4"
 
 #include "conf.h"
 #include "privs.h"
@@ -40,6 +40,9 @@
 
 module sql_sqlite_module;
 
+#define SQL_SQLITE_START_FL_NOW		1
+#define SQL_SQLITE_START_FL_EXCL	2
+
 typedef struct db_conn_struct {
   char *dsn;
   char *user;
@@ -153,6 +156,85 @@ static int exec_cb(void *n, int ncols, char **cols,
   return 0;
 }
 
+static int exec_stmt(cmd_rec *cmd, db_conn_t *conn, char *stmt, char **errstr) {
+  int res;
+  char *ptr = NULL;
+  unsigned int nretries = 0;
+
+  PRIVS_ROOT
+  res = sqlite3_exec(conn->dbh, stmt, exec_cb, cmd, &ptr);
+  PRIVS_RELINQUISH
+
+  while (res != SQLITE_OK) {
+    if (res == SQLITE_BUSY) {
+      struct timeval tv;
+
+      sqlite3_free(ptr);
+
+      nretries++;
+      sql_log(DEBUG_FUNC, "attempt #%u, database busy, trying '%s' again",
+        nretries, stmt);
+
+      /* Sleep for short bit, then try again. */
+      tv.tv_sec = 0;
+      tv.tv_usec = 500000L;
+
+      if (select(0, NULL, NULL, NULL, &tv) < 0) {
+        if (errno == EINTR) {
+          pr_signals_handle();
+        }
+      }
+
+      PRIVS_ROOT
+      res = sqlite3_exec(conn->dbh, stmt, exec_cb, cmd, &ptr);
+      PRIVS_RELINQUISH
+
+      continue;
+    }
+
+    *errstr = pstrdup(cmd->pool, ptr);
+    sqlite3_free(ptr);
+
+    sql_log(DEBUG_FUNC, "error executing '%s': (%d) %s", stmt, res, errstr);
+    return -1;
+  }
+
+  if (ptr)
+    sqlite3_free(ptr);
+
+  return 0;
+}
+
+static int query_start(cmd_rec *cmd, db_conn_t *conn, int flags,
+    char **errstr) {
+  char *start_txn = NULL;
+
+  switch (flags) {
+    case SQL_SQLITE_START_FL_NOW:
+      start_txn = pstrdup(cmd->tmp_pool, "BEGIN IMMEDIATE");
+      break;
+
+    case SQL_SQLITE_START_FL_EXCL:
+      start_txn = pstrdup(cmd->tmp_pool, "BEGIN EXCLUSIVE");
+      break;
+
+    default:
+      start_txn = pstrdup(cmd->tmp_pool, "BEGIN");
+      break;
+  }
+
+  return exec_stmt(cmd, conn, start_txn, errstr);
+}
+
+static int query_run(cmd_rec *cmd, db_conn_t *conn, char *query,
+    char **errstr) {
+  return exec_stmt(cmd, conn, query, errstr);
+}
+
+static int query_finish(cmd_rec *cmd, db_conn_t *conn, char **errstr) {
+  return exec_stmt(cmd, conn, pstrdup(cmd->tmp_pool, "COMMIT"), errstr);
+}
+
 static modret_t *sql_sqlite_get_data(cmd_rec *cmd) {
   register unsigned int i;
   unsigned int count, k = 0;
@@ -235,17 +317,40 @@ MODRET sql_sqlite_open(cmd_rec *cmd) {
       conn->dsn, errstr);
 
     sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_open");
+
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
   }
 
+  /* Tell SQLite to only use in-memory journals.  This is necessary for
+   * mod_sql_sqlite to work properly, for SQLLog statements, when a chroot
+   * is used.  Note that the MEMORY journal mode of SQLite is supported
+   * only for SQLite-3.6.5 and later.
+   */
+  res = sqlite3_exec(conn->dbh, "PRAGMA journal_mode = MEMORY;", NULL, NULL,
+    NULL);
+  if (res != SQLITE_OK) {
+    sql_log(DEBUG_FUNC, "error setting MEMORY journal mode: %s",
+      sqlite3_errmsg(conn->dbh));
+  }
+
   /* Add some SQLite information to the logs. */
   sql_log(DEBUG_INFO, MOD_SQL_SQLITE_VERSION ": SQLite version: %s",
     sqlite3_libversion());
 
   entry->nconn++;
 
-  /* Set up our timer, if necessary. */
-  if (entry->ttl > 0) {
+  if (pr_sql_conn_policy == SQL_CONN_POLICY_PERSESSION) {
+    /* If the connection policy is PERSESSION... */
+    if (entry->nconn == 1) {
+      /* ...and we are actually opening the first connection to the database;
+       * we want to make sure this connection stays open, after this first use
+       * (as per Bug#3290).  To do this, we re-bump the connection count.
+       */
+      entry->nconn++;
+    }
+
+  } else if (entry->ttl > 0) {
+    /* Set up our timer, if necessary. */
     entry->timer = pr_timer_add(entry->ttl, -1, &sql_sqlite_module,
       sql_sqlite_timer_cb, "sqlite connection ttl");
 
@@ -376,10 +481,15 @@ MODRET sql_sqlite_def_conn(cmd_rec *cmd) {
       "named connection already exists");
   }
 
-  entry->ttl = (cmd->argc == 5) ? 
-    (int) strtol(cmd->argv[4], (char **) NULL, 10) : 0;
-  if (entry->ttl < 0) 
-    entry->ttl = 0;
+  if (cmd->argc == 5) {
+    entry->ttl = (int) strtol(cmd->argv[4], (char **) NULL, 10);
+    if (entry->ttl >= 1) {
+      pr_sql_conn_policy = SQL_CONN_POLICY_TIMER;
+
+    } else {
+      entry->ttl = 0;
+    }
+  }
 
   entry->timer = 0;
   entry->nconn = 0;
@@ -413,11 +523,10 @@ MODRET sql_sqlite_exit(cmd_rec *cmd) {
 }
 
 MODRET sql_sqlite_select(cmd_rec *cmd) {
-  int res;
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *mr = NULL;
-  char *query = NULL, *tmp = NULL;
+  char *errstr = NULL, *query = NULL;
   cmd_rec *close_cmd;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tsqlite cmd_select");
@@ -472,7 +581,7 @@ MODRET sql_sqlite_select(cmd_rec *cmd) {
       }
     }
 
-    query = pstrcat(cmd->tmp_pool, "SELECT ", query, NULL);    
+    query = pstrcat(cmd->tmp_pool, "SELECT ", query, NULL);
   }
 
   /* Log the query string */
@@ -481,32 +590,8 @@ MODRET sql_sqlite_select(cmd_rec *cmd) {
   /* Perform the query.  If it doesn't work, log the error, close the
    * connection, then return the error from the query processing.
    */
-  res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-  while (res != SQLITE_OK) {
-    char *errstr;
-
-    if (res == SQLITE_BUSY) {
-      struct timeval tv;
-
-      /* Sleep for short bit, then try again. */
-      tv.tv_sec = 0;
-      tv.tv_usec = 500000L;
-
-      if (select(0, NULL, NULL, NULL, &tv) < 0) {
-        if (errno == EINTR) {
-          pr_signals_handle();
-        }
-      }
-
-      res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-      continue;
-    }
-
-    errstr = pstrdup(cmd->pool, tmp);
-    sqlite3_free(tmp);
-
-    sql_log(DEBUG_FUNC, "error executing '%s': %s", query, errstr);
 
+  if (query_start(cmd, conn, 0, &errstr) < 0) {
     close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
     sql_sqlite_close(close_cmd);
     destroy_pool(close_cmd->pool);
@@ -515,8 +600,23 @@ MODRET sql_sqlite_select(cmd_rec *cmd) {
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
   }
 
-  if (tmp)
-    sqlite3_free(tmp);
+  if (query_run(cmd, conn, query, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_select");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
+
+  if (query_finish(cmd, conn, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_select");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
 
   mr = sql_sqlite_get_data(cmd);
   
@@ -530,16 +630,16 @@ MODRET sql_sqlite_select(cmd_rec *cmd) {
 }
 
 MODRET sql_sqlite_insert(cmd_rec *cmd) {
-  int res;
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *mr = NULL;
-  char *query = NULL, *tmp = NULL;
+  char *errstr = NULL, *query = NULL;
   cmd_rec *close_cmd;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tsqlite cmd_insert");
 
-  if (cmd->argc != 2 && cmd->argc != 4) {
+  if (cmd->argc != 2 &&
+      cmd->argc != 4) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_insert");
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, "badly formed request");
   }
@@ -560,13 +660,14 @@ MODRET sql_sqlite_insert(cmd_rec *cmd) {
     return mr;
   }
 
-  /* Construct the query string */
-  if (cmd->argc == 2)
+  /* Construct the query string. */
+  if (cmd->argc == 2) {
     query = pstrcat(cmd->tmp_pool, "INSERT ", cmd->argv[1], NULL);
 
-  else
-    query = pstrcat(cmd->tmp_pool, "INSERT INTO ", cmd->argv[1], " (",
-      cmd->argv[2], ") VALUES (", cmd->argv[3], ")", NULL);
+  } else {
+    query = pstrcat(cmd->tmp_pool, "INSERT INTO ", cmd->argv[1],
+      " (", cmd->argv[2], ") VALUES (", cmd->argv[3], ")", NULL);
+  }
 
   /* Log the query string */
   sql_log(DEBUG_INFO, "query \"%s\"", query);
@@ -575,42 +676,33 @@ MODRET sql_sqlite_insert(cmd_rec *cmd) {
    * connection (and log any errors there, too) then return the error
    * from the query processing.
    */
-  res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-  while (res != SQLITE_OK) {
-    char *errstr;
-
-    if (res == SQLITE_BUSY) {
-      struct timeval tv;
-
-      /* Sleep for short bit, then try again. */
-      tv.tv_sec = 0;
-      tv.tv_usec = 500000L;
-
-      if (select(0, NULL, NULL, NULL, &tv) < 0) {
-        if (errno == EINTR) {
-          pr_signals_handle();
-        }
-      }
-
-      res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-      continue;
-    }
-
-    errstr = pstrdup(cmd->pool, tmp);
-    sqlite3_free(tmp);
-
-    sql_log(DEBUG_FUNC, "error executing '%s': %s", query, errstr);
 
+  if (query_start(cmd, conn, SQL_SQLITE_START_FL_NOW, &errstr) < 0) {
     close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
     sql_sqlite_close(close_cmd);
     destroy_pool(close_cmd->pool);
-    
+
     sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_insert");
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
   }
 
-  if (tmp)
-    sqlite3_free(tmp);
+  if (query_run(cmd, conn, query, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_insert");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
+
+  if (query_finish(cmd, conn, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_insert");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
 
   /* Reset these variables.  The memory in them is allocated from this
    * same cmd_rec, and will be recovered when the cmd_rec is destroyed.
@@ -628,11 +720,10 @@ MODRET sql_sqlite_insert(cmd_rec *cmd) {
 }
 
 MODRET sql_sqlite_update(cmd_rec *cmd) {
-  int res;
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *mr = NULL;
-  char *query = NULL, *tmp = NULL;
+  char *errstr = NULL, *query = NULL;
   cmd_rec *close_cmd;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tsqlite cmd_update");
@@ -658,16 +749,18 @@ MODRET sql_sqlite_update(cmd_rec *cmd) {
     return mr;
   }
 
+  /* Construct the query string. */
   if (cmd->argc == 2) {
     query = pstrcat(cmd->tmp_pool, "UPDATE ", cmd->argv[1], NULL);
 
   } else {
-    /* Construct the query string */
     query = pstrcat(cmd->tmp_pool, "UPDATE ", cmd->argv[1], " SET ",
       cmd->argv[2], NULL);
 
-    if (cmd->argc > 3 && cmd->argv[3])
+    if (cmd->argc > 3 &&
+        cmd->argv[3]) {
       query = pstrcat(cmd->tmp_pool, query, " WHERE ", cmd->argv[3], NULL);
+    }
   }
 
   /* Log the query string. */
@@ -676,42 +769,33 @@ MODRET sql_sqlite_update(cmd_rec *cmd) {
   /* Perform the query.  If it doesn't work close the connection, then
    * return the error from the query processing.
    */
-  res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-  while (res != SQLITE_OK) {
-    char *errstr;
-
-    if (res == SQLITE_BUSY) {
-      struct timeval tv;
-
-      /* Sleep for short bit, then try again. */
-      tv.tv_sec = 0;
-      tv.tv_usec = 500000L;
-
-      if (select(0, NULL, NULL, NULL, &tv) < 0) {
-        if (errno == EINTR) {
-          pr_signals_handle();
-        }
-      }
-
-      res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-      continue;
-    }
-
-    errstr = pstrdup(cmd->pool, tmp);
-    sqlite3_free(tmp);
-
-    sql_log(DEBUG_FUNC, "error executing '%s': %s", query, errstr);
 
+  if (query_start(cmd, conn, SQL_SQLITE_START_FL_NOW, &errstr) < 0) {
     close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
     sql_sqlite_close(close_cmd);
     destroy_pool(close_cmd->pool);
-   
+
     sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_update");
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
   }
 
-  if (tmp)
-    sqlite3_free(tmp);
+  if (query_run(cmd, conn, query, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_update");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
+
+  if (query_finish(cmd, conn, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_update");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
 
   /* Reset these variables.  The memory in them is allocated from this
    * same cmd_rec, and will be recovered when the cmd_rec is destroyed.
@@ -734,7 +818,11 @@ MODRET sql_sqlite_prepare(cmd_rec *cmd) {
   }
 
   conn_pool = (pool *) cmd->argv[0];
-  conn_cache = make_array(conn_pool, DEF_CONN_POOL_SIZE, sizeof(conn_entry_t));
+
+  if (conn_cache == NULL) {
+    conn_cache = make_array(conn_pool, DEF_CONN_POOL_SIZE,
+      sizeof(conn_entry_t *));
+  }
 
   return mod_create_data(cmd, NULL);
 }
@@ -753,11 +841,10 @@ MODRET sql_sqlite_procedure(cmd_rec *cmd) {
 }
 
 MODRET sql_sqlite_query(cmd_rec *cmd) {
-  int res;
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *mr = NULL;
-  char *query = NULL, *tmp = NULL;
+  char *errstr = NULL, *query = NULL;
   cmd_rec *close_cmd;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tsqlite cmd_query");
@@ -783,7 +870,7 @@ MODRET sql_sqlite_query(cmd_rec *cmd) {
     return mr;
   }
 
-  query = pstrcat(cmd->tmp_pool, cmd->argv[1], NULL);
+  query = pstrdup(cmd->tmp_pool, cmd->argv[1]);
 
   /* Log the query string */
   sql_log(DEBUG_INFO, "query \"%s\"", query);
@@ -791,42 +878,33 @@ MODRET sql_sqlite_query(cmd_rec *cmd) {
   /* Perform the query.  If it doesn't work close the connection, then
    * return the error from the query processing.
    */
-  res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-  while (res != SQLITE_OK) {
-    char *errstr;
-
-    if (res == SQLITE_BUSY) {
-      struct timeval tv;
-
-      /* Sleep for short bit, then try again. */
-      tv.tv_sec = 0;
-      tv.tv_usec = 500000L;
-
-      if (select(0, NULL, NULL, NULL, &tv) < 0) {
-        if (errno == EINTR) {
-          pr_signals_handle();
-        }
-      }
-
-      res = sqlite3_exec(conn->dbh, query, exec_cb, cmd, &tmp);
-      continue;
-    }
-
-    errstr = pstrdup(cmd->pool, tmp);
-    sqlite3_free(tmp);
-
-    sql_log(DEBUG_FUNC, "error executing '%s': %s", query, errstr);
 
+  if (query_start(cmd, conn, 0, &errstr) < 0) {
     close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
     sql_sqlite_close(close_cmd);
     destroy_pool(close_cmd->pool);
-  
+
     sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_query");
     return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
   }
 
-  if (tmp)
-    sqlite3_free(tmp);
+  if (query_run(cmd, conn, query, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_query");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
+
+  if (query_finish(cmd, conn, &errstr) < 0) {
+    close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+    sql_sqlite_close(close_cmd);
+    destroy_pool(close_cmd->pool);
+
+    sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_query");
+    return PR_ERROR_MSG(cmd, MOD_SQL_SQLITE_VERSION, errstr);
+  }
 
   mr = sql_sqlite_get_data(cmd);
   
@@ -846,6 +924,7 @@ MODRET sql_sqlite_quote(cmd_rec *cmd) {
   char *unescaped = NULL;
   char *escaped = NULL;
   char *tmp;
+  cmd_rec *close_cmd;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tsqlite cmd_escapestring");
 
@@ -876,6 +955,10 @@ MODRET sql_sqlite_quote(cmd_rec *cmd) {
   escaped = pstrdup(cmd->pool, tmp);
   sqlite3_free(tmp);
 
+  close_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, entry->name);
+  sql_sqlite_close(close_cmd);
+  destroy_pool(close_cmd->pool);
+
   sql_log(DEBUG_FUNC, "%s", "exiting \tsqlite cmd_escapestring");
   return mod_create_data(cmd, escaped);
 }
@@ -989,9 +1072,15 @@ static int sql_sqlite_init(void) {
 }
 
 static int sql_sqlite_sess_init(void) {
-  conn_pool = make_sub_pool(session.pool);
-  conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
-    sizeof(conn_entry_t));
+  if (conn_pool == NULL) {
+    conn_pool = make_sub_pool(session.pool);
+    pr_pool_tag(conn_pool, "SQLite connection pool");
+  }
+
+  if (conn_cache == NULL) {
+    conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
+      sizeof(conn_entry_t *));
+  }
 
   return 0;
 }
