@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2008 The ProFTPD Project team
+ * Copyright (c) 2001-2010 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Data connection management functions
- * $Id: data.c,v 1.116 2008/12/03 05:06:14 castaglia Exp $
+ * $Id: data.c,v 1.120.2.3 2010/09/01 20:46:00 castaglia Exp $
  */
 
 #include "conf.h"
@@ -337,7 +337,7 @@ static int data_active_open(char *reason, off_t size) {
   if (!reason && session.xfer.filename)
     reason = session.xfer.filename;
 
-  session.d = pr_inet_create_connection(session.pool, NULL, -1,
+  session.d = pr_inet_create_conn(session.pool, NULL, -1,
     session.c->local_addr, session.c->local_port-1, TRUE);
 
   /* Set the "stalled" timer, if any, to prevent the connection
@@ -632,7 +632,7 @@ void pr_data_close(int quiet) {
 
   session.sf_flags &= (SF_ALL^SF_PASSIVE);
   session.sf_flags &= (SF_ALL^(SF_ABORT|SF_XFER|SF_PASSIVE|SF_ASCII_OVERRIDE));
-  session_set_idle();
+  pr_session_set_idle();
 
   if (!quiet)
     pr_response_add(R_226, _("Transfer complete"));
@@ -683,7 +683,7 @@ void pr_data_abort(int err, int quiet) {
 
   session.sf_flags &= (SF_ALL^SF_PASSIVE);
   session.sf_flags &= (SF_ALL^(SF_XFER|SF_PASSIVE|SF_ASCII_OVERRIDE));
-  session_set_idle();
+  pr_session_set_idle();
 
   /* Aborts no longer necessary */
   signal(SIGURG, SIG_IGN);
@@ -894,6 +894,11 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
        * already have a data transfer in progress.  For any data transfer
        * command, send a 450 ("busy") reply.  Looks like almost all of the
        * data transfer commands accept that response, as per RFC959.
+       *
+       * We also prevent the EPRT, EPSV, PASV, and PORT commands, since
+       * they will also interfere with the current data transfer.  In doing
+       * so, we break RFC compliance a little; RFC959 does not allow a
+       * response code of 450 for those commands (although it should).
        */
       if (strcmp(cmd->argv[0], C_APPE) == 0 ||
           strcmp(cmd->argv[0], C_LIST) == 0 ||
@@ -901,7 +906,13 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
           strcmp(cmd->argv[0], C_NLST) == 0 ||
           strcmp(cmd->argv[0], C_RETR) == 0 ||
           strcmp(cmd->argv[0], C_STOR) == 0 ||
-          strcmp(cmd->argv[0], C_STOU) == 0) {
+          strcmp(cmd->argv[0], C_STOU) == 0 ||
+          strcmp(cmd->argv[0], C_RNFR) == 0 ||
+          strcmp(cmd->argv[0], C_RNTO) == 0 ||
+          strcmp(cmd->argv[0], C_PORT) == 0 ||
+          strcmp(cmd->argv[0], C_EPRT) == 0 ||
+          strcmp(cmd->argv[0], C_PASV) == 0 ||
+          strcmp(cmd->argv[0], C_EPSV) == 0) {
         pool *resp_pool;
 
         pr_trace_msg(trace_channel, 5,
@@ -999,9 +1010,14 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
 #else
     xerrno = EIO;
 #endif
+
     pr_trace_msg(trace_channel, 1,
       "data connection is null prior to data transfer (possibly from "
       "aborted transfer), returning '%s' error", strerror(xerrno));
+    pr_log_debug(DEBUG5,
+      "data connection is null prior to data transfer (possibly from "
+       "aborted transfer), returning '%s' error", strerror(xerrno));
+
     errno = xerrno;
     return -1;
   }
@@ -1016,6 +1032,8 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
         buflen = session.xfer.buflen;        /* how much remains in buf */
         adjlen = 0;
 
+        pr_signals_handle();
+
         len = pr_netio_read(session.d->instrm, buf + buflen,
           session.xfer.bufsize - buflen, 1);
         if (len < 0)
@@ -1100,6 +1118,8 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
       int buflen = cl_size;
       unsigned int xferbuflen;
 
+      pr_signals_handle();
+
       if (buflen > pr_config_get_xfer_bufsz())
         buflen = pr_config_get_xfer_bufsz();
 
@@ -1142,6 +1162,13 @@ int pr_data_xfer(char *cl_buf, int cl_size) {
 
   session.xfer.total_bytes += total;
   session.total_bytes += total;
+  if (session.xfer.direction == PR_NETIO_IO_RD) {
+    session.total_bytes_in += total;
+
+  } else {
+    session.total_bytes_out += total;
+  }
+
   return (len < 0 ? -1 : len);
 }
 
@@ -1219,6 +1246,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
 
         session.xfer.total_bytes += len;
         session.total_bytes += len;
+        session.total_bytes_out += len;
 
         return -1;
       }
@@ -1236,6 +1264,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
 
       session.xfer.total_bytes += len;
       session.total_bytes += len;
+      session.total_bytes_out += len;
       total += len;
 
       pr_signals_handle();
@@ -1327,6 +1356,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
         if (XFER_ABORTED) {
           session.xfer.total_bytes += len;
           session.total_bytes += len;
+          session.total_bytes_out += len;
 
           return -1;
         }
@@ -1334,11 +1364,12 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
         pr_signals_handle();
 
         /* If we got everything in this transaction, we're done. */
-        if (len >= count)
+        if (len >= count) {
           break;
 
-        else 
+        } else {
           count -= len;
+        }
 
 	*offset += len;
 	
@@ -1350,6 +1381,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
 	
 	session.xfer.total_bytes += len;
 	session.total_bytes += len;
+	session.total_bytes_out += len;
 	total += len;
 	
 	continue;
@@ -1376,6 +1408,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
 
   session.xfer.total_bytes += len;
   session.total_bytes += len;
+  session.total_bytes_out += len;
   total += len;
 
   return total;
