@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2008 The ProFTPD Project team
+ * Copyright (c) 2001-2009 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,14 +25,15 @@
  */
 
 /* ProFTPD logging support.
- * $Id: log.c,v 1.89 2008/12/16 23:43:55 castaglia Exp $
+ * $Id: log.c,v 1.99.2.1 2010/04/13 21:57:34 castaglia Exp $
  */
 
 #include "conf.h"
 
 #include <signal.h>
 
-#define LOGBUFFER_SIZE	2048
+/* Max path length plus 64 bytes for additional info. */
+#define LOGBUFFER_SIZE		(PR_TUNABLE_PATH_MAX + 64)
 
 static int syslog_open = FALSE;
 static int syslog_discard = FALSE;
@@ -46,150 +47,23 @@ static int systemlog_fd = -1;
 
 int syslog_sockfd = -1;
 
-/* This next function logs an entry to wtmp, it MUST be called as
- * root BEFORE a chroot occurs.
- * Note: This has some portability ifdefs in it.  They *should* work,
- * but I haven't been able to test them.
- */
+#ifdef PR_USE_NONBLOCKING_LOG_OPEN
+static int fd_set_block(int fd) {
+  int flags, res;
 
-int log_wtmp(char *line, const char *name, const char *host,
-    pr_netaddr_t *ip) {
-  struct stat buf;
-  struct utmp ut;
-  int res = 0;
-  static int fd = -1;
-
-#if ((defined(SVR4) || defined(__SVR4)) || \
-    (defined(__NetBSD__) && defined(HAVE_UTMPX_H))) && \
-    !(defined(LINUX) || defined(__hpux) || defined (_AIX))
-  /* This "auxilliary" utmp doesn't exist under linux. */
-#if defined(__sparcv9) && !defined(__NetBSD__)
-  struct futmpx utx;
-  time_t t;
-#else
-  struct utmpx utx;
-#endif
-  static int fdx = -1;
-
-#if !defined(WTMPX_FILE) && defined(_PATH_WTMPX)
-# define WTMPX_FILE _PATH_WTMPX
-#endif
-
-  if (fdx < 0 &&
-      (fdx = open(WTMPX_FILE, O_WRONLY|O_APPEND, 0)) < 0) {
-    pr_log_pri(PR_LOG_WARNING, "wtmpx %s: %s", WTMPX_FILE, strerror(errno));
-    return -1;
-  }
-
-  /* Unfortunately, utmp string fields are terminated by '\0' if they are
-   * shorter than the size of the field, but if they are exactly the size of
-   * the field they don't have to be terminated at all.  Frankly, this sucks.
-   * Insane if you ask me.  Unless there's massive uproar, I prefer to err on
-   * the side of caution and always null-terminate our strings.
-   */
-  if (fstat(fdx, &buf) == 0) {
-    memset(&utx, 0, sizeof(utx));
-
-    sstrncpy(utx.ut_user, name, sizeof(utx.ut_user));
-    sstrncpy(utx.ut_id, "ftp", sizeof(utx.ut_user));
-    sstrncpy(utx.ut_line, line, sizeof(utx.ut_line));
-    sstrncpy(utx.ut_host, host, sizeof(utx.ut_host));
-    utx.ut_pid = session.pid ? session.pid : getpid();
-#if defined(__NetBSD__) && defined(HAVE_UTMPX_H)
-    memcpy(&utx.ut_ss, pr_netaddr_get_inaddr(ip), sizeof(utx.ut_ss));
-    gettimeofday(&utx.ut_tv, NULL);
-#else /* SVR4 */
-    utx.ut_syslen = strlen(utx.ut_host)+1;
-#  ifdef __sparcv9
-    time(&t);
-    utx.ut_tv.tv_sec = (time32_t)t;
-#  else
-    time(&utx.ut_tv.tv_sec);
-#  endif
-#endif /* SVR4 */
-
-    if (*name)
-      utx.ut_type = USER_PROCESS;
-    else
-      utx.ut_type = DEAD_PROCESS;
-#ifdef HAVE_UT_UT_EXIT
-    utx.ut_exit.e_termination = 0;
-    utx.ut_exit.e_exit = 0;
-#endif /* HAVE_UT_UT_EXIT */
-    if (write(fdx, (char *)&utx, sizeof(utx)) != sizeof(utx))
-      (void) ftruncate(fdx, buf.st_size);
-
-  } else {
-    pr_log_debug(DEBUG0, "%s fstat(): %s", WTMPX_FILE, strerror(errno));
-    res = -1;
-  }
-
-#else /* Non-SVR4 systems */
-
-  if (fd < 0 &&
-      (fd = open(WTMP_FILE, O_WRONLY|O_APPEND, 0)) < 0) {
-    pr_log_pri(PR_LOG_WARNING, "wtmp %s: %s", WTMP_FILE, strerror(errno));
-    return -1;
-  }
-
-  if (fstat(fd, &buf) == 0) {
-    memset(&ut, 0, sizeof(ut));
-#ifdef HAVE_UTMAXTYPE
-# ifdef LINUX
-    if (ip)
-#  ifndef PR_USE_IPV6
-      memcpy(&ut.ut_addr, pr_netaddr_get_inaddr(ip), sizeof(ut.ut_addr));
-#  else
-      memcpy(&ut.ut_addr_v6, pr_netaddr_get_inaddr(ip), sizeof(ut.ut_addr_v6));
-#  endif /* !PR_USE_IPV6 */
-# else
-    sstrncpy(ut.ut_id, "ftp", sizeof(ut.ut_id));
-#  ifdef HAVE_UT_UT_EXIT
-    ut.ut_exit.e_termination = 0;
-    ut.ut_exit.e_exit = 0;
-#  endif /* !HAVE_UT_UT_EXIT */
-# endif /* !LINUX */
-    sstrncpy(ut.ut_line, line, sizeof(ut.ut_line));
-    if (name && *name)
-      sstrncpy(ut.ut_user, name, sizeof(ut.ut_user));
-    ut.ut_pid = session.pid ? session.pid : getpid();
-    if (name && *name)
-      ut.ut_type = USER_PROCESS;
-    else
-      ut.ut_type = DEAD_PROCESS;
-#else  /* !HAVE_UTMAXTYPE */
-    sstrncpy(ut.ut_line, line, sizeof(ut.ut_line));
-    if (name && *name)
-      sstrncpy(ut.ut_name, name, sizeof(ut.ut_name));
-#endif /* HAVE_UTMAXTYPE */
-
-#ifdef HAVE_UT_UT_HOST
-    if (host && *host)
-      sstrncpy(ut.ut_host, host, sizeof(ut.ut_host));
-#endif /* HAVE_UT_UT_HOST */
-
-    ut.ut_time = time(NULL);
-    if (write(fd, (char *)&ut, sizeof(ut)) != sizeof(ut)) {
-      if (ftruncate(fd, buf.st_size) < 0) {
-        pr_log_debug(DEBUG0, "error truncating '%s': %s", WTMP_FILE,
-          strerror(errno));
-      }
-    }
-
-  } else {
-    pr_log_debug(DEBUG0, "%s fstat(): %s", WTMP_FILE, strerror(errno));
-    res = -1;
-  }
-#endif /* SVR4 */
+  flags = fcntl(fd, F_GETFL);
+  res = fcntl(fd, F_SETFL, flags & (U32BITS ^ O_NONBLOCK));
 
   return res;
 }
+#endif /* PR_USE_NONBLOCKING_LOG_OPEN */
 
 int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
+  int res;
   pool *tmp_pool = NULL;
   char *tmp = NULL, *lf;
   unsigned char have_stat = FALSE, *allow_log_symlinks = NULL;
-  struct stat sbuf;
+  struct stat st;
 
   /* Sanity check */
   if (!log_file || !log_fd) {
@@ -206,6 +80,8 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   if (tmp == NULL) {
     pr_log_debug(DEBUG0, "inappropriate log file: %s", lf);
     destroy_pool(tmp_pool);
+
+    errno = EINVAL;
     return -1;
   }
 
@@ -214,22 +90,27 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
    */
   *tmp = '\0';
 
-  if (stat(lf, &sbuf) == -1) {
+  if (stat(lf, &st) < 0) {
+    int xerrno = errno;
     pr_log_debug(DEBUG0, "error: unable to stat() %s: %s", lf,
       strerror(errno));
     destroy_pool(tmp_pool);
+
+    errno = xerrno;
     return -1;
   }
 
   /* The path must be in a valid directory */
-  if (!S_ISDIR(sbuf.st_mode)) {
+  if (!S_ISDIR(st.st_mode)) {
     pr_log_debug(DEBUG0, "error: %s is not a directory", lf);
     destroy_pool(tmp_pool);
+
+    errno = ENOTDIR;
     return -1;
   }
 
   /* Do not log to world-writeable directories */
-  if (sbuf.st_mode & S_IWOTH) {
+  if (st.st_mode & S_IWOTH) {
     pr_log_pri(PR_LOG_NOTICE, "error: %s is a world writeable directory", lf);
     destroy_pool(tmp_pool);
     return PR_LOG_WRITABLE_DIR;
@@ -243,9 +124,18 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   allow_log_symlinks = get_param_ptr(main_server->conf, "AllowLogSymlinks",
     FALSE);
 
-  if (!allow_log_symlinks || *allow_log_symlinks == FALSE) {
+  if (allow_log_symlinks == NULL ||
+      *allow_log_symlinks == FALSE) {
     int flags = O_APPEND|O_CREAT|O_WRONLY;
 
+#ifdef PR_USE_NONBLOCKING_LOG_OPEN
+    /* Use the O_NONBLOCK flag when opening log files, as they might be
+     * FIFOs whose other end is not currently running; we do not want to
+     * block indefinitely in such cases.
+     */
+    flags |= O_NONBLOCK;
+#endif /* PR_USE_NONBLOCKING_LOG_OPEN */
+
 #ifdef O_NOFOLLOW
     /* On systems that support the O_NOFOLLOW flag (e.g. Linux and FreeBSD),
      * use it so that the path being opened, if it is a symlink, is not
@@ -265,7 +155,7 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
 #endif /* O_NOFOLLOW or SOLARIS2 */
 
     *log_fd = open(lf, flags, log_mode);
-    if (*log_fd == -1) {
+    if (*log_fd < 0) {
 
       if (errno != EEXIST) {
         destroy_pool(tmp_pool);
@@ -306,7 +196,7 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
         flags &= ~O_EXCL;
 
         *log_fd = open(lf, flags, log_mode);
-        if (*log_fd == -1) {
+        if (*log_fd < 0) {
           destroy_pool(tmp_pool);
           return -1;
         }
@@ -315,7 +205,7 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
          * above and the lstat() call below...
          */
 
-        if (lstat(lf, &sbuf) != -1)
+        if (lstat(lf, &st) != -1)
           have_stat = TRUE;
 #else
         destroy_pool(tmp_pool);
@@ -325,10 +215,12 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
     }
 
     /* Stat the file using the descriptor, not the path */
-    if (!have_stat && fstat(*log_fd, &sbuf) != -1)
+    if (!have_stat &&
+        fstat(*log_fd, &st) != -1)
       have_stat = TRUE;
 
-    if (!have_stat || S_ISLNK(sbuf.st_mode)) {
+    if (!have_stat ||
+        S_ISLNK(st.st_mode)) {
       pr_log_debug(DEBUG0, !have_stat ? "error: unable to stat %s" :
         "error: %s is a symbolic link", lf);
 
@@ -339,22 +231,51 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
     }
 
   } else {
-    *log_fd = open(lf, O_CREAT|O_APPEND|O_WRONLY, log_mode);
-    if (*log_fd == -1) {
+    int flags = O_CREAT|O_APPEND|O_WRONLY;
+
+#ifdef PR_USE_NONBLOCKING_LOG_OPEN
+    /* Use the O_NONBLOCK flag when opening log files, as they might be
+     * FIFOs whose other end is not currently running; we do not want to
+     * block indefinitely in such cases.
+     */
+    flags |= O_NONBLOCK;
+#endif /* PR_USE_NONBLOCKING_LOG_OPEN */
+
+    *log_fd = open(lf, flags, log_mode);
+    if (*log_fd < 0) {
       destroy_pool(tmp_pool);
       return -1;
     }
   }
 
+  /* Find a usable fd for the just-opened log fd. */
+  if (*log_fd <= STDERR_FILENO) {
+    res = pr_fs_get_usable_fd(*log_fd);
+    if (res < 0) {
+      pr_log_debug(DEBUG0, "warning: unable to find good fd for logfd %d: %s",
+        *log_fd, strerror(errno));
+
+    } else {
+      close(*log_fd);
+      *log_fd = res;
+    }
+  }
+
+#ifdef PR_USE_NONBLOCKING_LOG_OPEN
+  /* Return the fd to blocking mode. */
+  fd_set_block(*log_fd);
+#endif /* PR_USE_NONBLOCKING_LOG_OPEN */
+
   destroy_pool(tmp_pool);
   return 0;
 }
 
-int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
+int pr_log_vwritefile(int logfd, const char *ident, const char *fmt,
+    va_list msg) {
   char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
   time_t timestamp = time(NULL);
   struct tm *t = NULL;
-  va_list msg;
+  size_t buflen;
 
   if (logfd < 0) {
     errno = EINVAL;
@@ -375,12 +296,17 @@ int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
   buf[sizeof(buf)-1] = '\0';
 
   /* Affix the message */
-  va_start(msg, fmt);
   vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), fmt, msg);
-  va_end(msg);
 
   buf[sizeof(buf)-1] = '\0';
-  buf[strlen(buf)] = '\n';
+
+  buflen = strlen(buf);
+  if (buflen < (sizeof(buf) - 1)) {
+    buf[buflen] = '\n';
+
+  } else {
+    buf[sizeof(buf)-2] = '\n';
+  }
 
   while (write(logfd, buf, strlen(buf)) < 0) {
     if (errno == EINTR) {
@@ -394,6 +320,22 @@ int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
   return 0;
 }
 
+int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
+  va_list msg;
+  int res;
+
+  if (logfd < 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  va_start(msg, fmt);
+  res = pr_log_vwritefile(logfd, ident, fmt, msg);
+  va_end(msg);
+
+  return res;
+}
+
 int log_opensyslog(const char *fn) {
   int res = 0;
 
@@ -436,6 +378,10 @@ void log_closesyslog(void) {
   syslog_open = FALSE;
 }
 
+int log_getfacility(void) {
+  return set_facility;
+}
+
 void log_setfacility(int f) {
   set_facility = f;
 }
